#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

struct job_request {
    pthread_cond_t  cond;
    pthread_mutex_t  lock;
    int     work;
    struct job_request *next;
};

pthread_cond_t  job_queue_cond;
pthread_mutex_t job_queue_lock;
struct job_request *job_requests;

void *worker_thread(void *arg) {
    struct job_request job;

    pthread_cond_init(&job.cond, NULL);
    pthread_mutex_init(&job.lock, NULL);

    // Insert a job request into the job request queue.
    pthread_mutex_lock(&job.lock);
    job.work = -1;
    while (job.work != 0) {
        pthread_mutex_lock(&job_queue_lock);
        job.next = job_requests;
        job.work = -1;
        job_requests = &job;
        pthread_mutex_unlock(&job_queue_lock);
        pthread_cond_broadcast(&job_queue_cond);
        pthread_cond_wait(&job.cond, &job.lock);
        if (job.work != -1)
            printf("Got job %d worker: %d\n", job.work, (int)arg);

    }
    return NULL;
}

void job_dispatcher(int jobs, int thread_count) {
    int i = 0;
    printf("jobs=%d\n", jobs);
    pthread_mutex_lock(&job_queue_lock);
    while (i < (jobs + thread_count)) {
        if (!job_requests)

            pthread_cond_wait(&job_queue_cond, &job_queue_lock);
        
        while (job_requests) {
            struct job_request *job = job_requests;
            pthread_mutex_lock(&job->lock);
            job_requests = job_requests->next;
            if (i < jobs)
                job->work = i + 1;
            else
                job->work = 0;
            pthread_cond_signal(&job->cond);
            pthread_mutex_unlock(&job->lock);
            ++i;
        }
    }
    pthread_mutex_unlock(&job_queue_lock);
}

int main(int argc, char *argv[]) {
    if (argc != 3) {
        printf("usage: %s worker_threads jobs\n", argv[0]);
        return -__LINE__;
    }
    int num_threads = atoi(argv[1]);
    int jobs = atoi(argv[2]);
    int64_t i;

    pthread_cond_init(&job_queue_cond, NULL);
    pthread_mutex_init(&job_queue_lock, NULL);

    pthread_t *threads = malloc(sizeof(*threads) * num_threads);
    for (i = 0; i < num_threads; i++)
        pthread_create(&threads[i], NULL, worker_thread, (void *)i);

    job_dispatcher(jobs, num_threads);

    for (i = 0; i < num_threads; i++)
        pthread_join(threads[i], NULL);
}