#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

pthread_mutex_t     extra_libc_lock;

struct generate_work {
    int *start;
    int num_items;
};
void *generate_random_numbers(void *arg) {
    struct generate_work  *generate_work = arg;
    int i;

    for (i = 0; i < generate_work->num_items; i++) {
        pthread_mutex_lock(&extra_libc_lock);
        generate_work->start[i] = rand() % 10;
        pthread_mutex_unlock(&extra_libc_lock);
    }

    return NULL;
}


struct sum_work {
    int *start;
    int num_items;
    int *result;
    pthread_mutex_t   *result_lock;
};
void *sum_numbers(void *arg) {
    struct sum_work  *sum_work = arg;
    int i, res = 0;

    for (i = 0; i < sum_work->num_items; i++)
        res += sum_work->start[i];

    pthread_mutex_lock(sum_work->result_lock);
    *sum_work->result += res;
    pthread_mutex_unlock(sum_work->result_lock);

    return NULL;
}


int main(int argc, char *argv[]) {
    pthread_t   *threads;
    int         num_threads, i;

    if (argc != 3) {
        printf("usage: %s threads array_size\n", argv[0]);
        return -__LINE__;
    }

    int array_size = atoi(argv[2]);
    int *array = malloc(sizeof(*array) * array_size);
    num_threads = atoi(argv[1]);

    // Generate random numbers
    pthread_mutex_init(&extra_libc_lock, NULL);
    struct generate_work    *generate_work = malloc(sizeof(*generate_work) * num_threads);
    int chunk = array_size / num_threads;
    threads = malloc(sizeof(*threads) * num_threads);

    for (i = 0; i < num_threads; i++) {
        generate_work[i].start = &array[chunk * i];
        if ((i + 1) != num_threads)
            generate_work[i].num_items = chunk;
        else
            generate_work[i].num_items = (array_size - chunk * i);

        if (pthread_create(&threads[i], NULL, generate_random_numbers, &generate_work[i]) != 0) {
            perror("Failed to create thread\n");
        }
    }
    // Join threads
    for (i = 0; i < num_threads; i++) {
        pthread_join(threads[i], NULL);
    }
    free(generate_work);

    // Sum them up
    struct sum_work *sum_work = malloc(sizeof(*sum_work) * num_threads);
    int sum = 0;
    pthread_mutex_t sum_lock;
    pthread_mutex_init(&sum_lock, NULL);

    for (i = 0; i < num_threads; i++) {
        sum_work[i].start = &array[chunk * i];
        if ((i + 1) != num_threads)
            sum_work[i].num_items = chunk;
        else
            sum_work[i].num_items = (array_size - chunk * i);
        sum_work[i].result = &sum;
        sum_work[i].result_lock = &sum_lock;
        if (pthread_create(&threads[i], NULL, sum_numbers, &sum_work[i]) != 0) {
            perror("Failed to create thread\n");
        }

    }
    // Join threads
    for (i = 0; i < num_threads; i++) {
        pthread_join(threads[i], NULL);
    }
    free(sum_work);
    free(threads);

    printf("Sum is: %d\n", sum);

    return 0;
}
